open Ast;;

(*********** Functions related to lazy list **********)

type 'a thunk = Thunk of (unit -> 'a) | Val of 'a
type 'a lzlist = Nil | Cons of 'a * ('a lzlist thunk)

let force v = match v with 
  | Thunk x -> x()
  | Val x -> x

let rec lzappend l1 l2 = match l1 with
  | Nil -> (force l2)
  | Cons(l11, l1s) -> Cons(l11, Thunk(fun () -> (lzappend (force l1s) l2)))

let rec lzmap (f: 'a -> 'b) (l: 'a lzlist): 'b lzlist = match l with
  | Nil -> Nil
  | Cons(l1, ls) -> Cons((f l1), Thunk(fun () -> lzmap f (force ls)))

let rec lzfilter (f: 'a -> bool) (l: 'a lzlist): 'a lzlist = match l with
  | Nil -> Nil
  | Cons(l1, ls) -> if (f l1) 
      then Cons(l1, Thunk(fun () -> lzfilter f (force ls)))
    else (lzfilter f (force ls))

let rec lzflatten (ll: 'a lzlist lzlist): 'a lzlist = match ll with
  | Nil -> Nil
  | Cons(ll1, lls) -> match ll1 with
    | Nil -> lzflatten (force lls)
    | Cons(l1, l1s) -> Cons(l1, 
                            Thunk(fun() -> lzflatten (Cons((force l1s), lls))))

(************* Functions related to parsing **************)

let rec recognize_vars (Nvar(n, tl)) map = 
  if ((Char.lowercase n.[0]) <> n.[0]) 
    then try
           let v = List.assoc n map in (Var(v), map)
      with Not_found ->
        let v = List.length map in (Var(v), (n, v)::map)
  else let (tl', map') = recognize_varlist tl map in
       (Nvar(n, tl'), map')
and
    recognize_varlist tl map = match tl with
    | [] -> ([], map)
    | t::ts -> let (t', map') = recognize_vars t map in
               let (ts', map'') = recognize_varlist ts map' in
               (t'::ts', map'')

let recognize_varlist1 tl: term list = 
  let (tl', _) = recognize_varlist tl [] in tl'

let parse_file f = 
  let chan = open_in f in
  let lexbuf = Lexing.from_channel chan in
  Parser.program Lexer.token lexbuf
;;

(************ Functions related to printing terms ************)

let rec print_list sep prtfn l =
  let doprt i elem = (if i > 0 then (print_string sep)); (prtfn elem)
  in (List.iteri doprt l)

let rec numvars t = match t with 
  | Var(i) -> 1 
  | Nvar(_,ts) -> numvarslist ts
and numvarslist ts = 
  List.fold_left (fun x y -> (x+y)) 0 (List.map numvars ts)

let rec prtList tl flag = match tl with
  | t1::[] -> print_string "|"; print_term t1; print_string "]"
  | t1::ts ->
      (if flag then print_string ", "); print_term t1; 
      match ts with
      | Nvar("[]", [])::[] -> print_string "]"
      | _ -> prtList ts true

and print_term t = 
  let rec flatten_list t' = match t' with
    | Var(x) -> [t']
    | Nvar(n, ts') -> if n = "|" 
        then let t1::t2::[] = ts' in t1::(flatten_list  t2)
      else [t']
  in match t with
  | Var(x) -> print_string "X"; print_int x
  | Nvar(n, ts) -> 
      if n = "|" then 
        let tl = flatten_list t in (print_string "["; prtList tl false)
      else if ts = [] 
        then (print_string n)
      else (print_string n; print_string "("; 
            (print_list ", " print_term ts); print_string ")")

let print_env (bp, subst) = 
  let print1_subst (v, t) = 
    (print_term (Var v)); print_string "="; (print_term t) 
  in print_string "{"; (print_list ", " print1_subst subst); print_string "}"

let print_envs ss = (print_list "\n" print_env ss)

let rec vars t = match t with
  | Var(i) -> [i]
  | Nvar(n, ts) -> (List.flatten (List.map vars ts))

let rec mk_full_term vl subst flag t = match t with
  | Var(x) -> (if  not flag && (List.mem x vl) then t
               else try 
                 let t' = List.assoc x subst in mk_full_term vl subst false t'
               with Not_found -> t)
  | Nvar(n, ts) -> if ts = [] then t 
                   else Nvar(n, (List.map (mk_full_term vl subst false) ts))

let print_with_subst vl envs = 
  let print_with_1subst (_, subst) = 
    let print_1var v = 
      print_term (Var v); print_string "="; 
      print_term (mk_full_term vl subst true (Var v))
    in if vl=[] then (print_string "yes") else (print_list ", " print_1var vl)
  in (print_list "\n" print_with_1subst envs)
;;

let rec print_with_lzsubst vl envs = 
  let print_with_1subst (_, subst) = 
    let print_1var v = 
      (print_term (Var v)); print_string "="; 
      print_term (mk_full_term vl subst true (Var v))
    in if vl=[] then (print_string "yes") else (print_list ", " print_1var vl)
  in match envs with 
  | Nil -> print_string "no\n"
  | Cons(env1, envrest) ->
    print_with_1subst env1; print_string "\n";
    if read_line() = ";"
      then print_with_lzsubst vl (force envrest)
